use core::time::Duration;
use std::collections::HashMap;

use bitfinex::api::Bitfinex;
use bitfinex::positions::Position;
use tokio::time::delay_for;

use crate::BoxError;
use crate::currency::{Symbol, SymbolPair};
use crate::pairs::PairStatus;
use crate::ticker::Ticker;
use bitfinex::ticker::TradingPairTicker;

pub struct BfxWrapper {
    bfx: Bitfinex
}

impl BfxWrapper {
    pub fn new(api_key: &str, api_secret: &str) -> Self {
        BfxWrapper {
            bfx: Bitfinex::new(Some(api_key.into()), Some(api_secret.into()))
        }
    }

    pub async fn current_prices(&self, pair: &SymbolPair) -> Result<TradingPairTicker, BoxError> {
        let ticker: TradingPairTicker = self.bfx.ticker.trading_pair(pair.clone()).await?;

        Ok(ticker)
    }
}

pub struct BfxBot {
    pub bfx: BfxWrapper,
    ticker: Ticker,
    pair_status: Vec<PairStatus>,
    quote: Symbol,
    trading_symbols: Vec<Symbol>,
    // account_info: String,
    // ledger: String,
}

impl BfxBot {
    pub fn new<S: Into<String>>(api_key: S, api_secret: S, trading_symbols: Vec<Symbol>, quote: Symbol, tick_duration: Duration) -> Self {
        BfxBot {
            bfx: BfxWrapper::new(&api_key.into(), &api_secret.into()),
            ticker: Ticker::new(tick_duration),
            pair_status: trading_symbols
                .iter()
                .map(|x| SymbolPair::new(quote.clone(), x.clone()))
                .map(|x| PairStatus::new(x, 1, None))
                .collect(),
            quote,
            // account_info: String::new(),
            // ledger: String::new(),
            trading_symbols,
        }
    }

    pub async fn current_prices(&self, symbol: Symbol) -> Result<TradingPairTicker, BoxError> {
        let trading_pair = SymbolPair::new(self.quote.clone(), symbol.clone());

        if !self.trading_symbols.contains(&symbol){
            return Err("Symbol not supported.".into());
        }

        self.bfx.current_prices(&trading_pair).await
    }

    pub async fn update(&mut self) {
        println!("Updating...");
        delay_for(self.ticker.duration()).await;
        self.ticker.inc();
        // self.update_pairs().await;
        println!("Done!");
    }

    // async fn update_pairs(&mut self) {
    //     let active_positions = self.bfx.positions.active_positions().await?;
    //
    //     for p in active_positions {}
    // }
}

